﻿Public Class TemperatureController

    Public Shared Active As Boolean = False
    Public Shared LoopTime As Double

    Public Enabled As Boolean = False
    Public SlotSensor As Int32
    Public SlotHeater As Int32
    Public DesiredTemp As Double
    Public DefaultTemp As Double
    Public Pgain As Double
    Public Igain As Double
    Public Dgain As Double

    Public Measured As Double
    Public Nvalues As Int32
    Public Table(,) As Double

    Private ErrorNew As Double
    Private ErrorOld As Double
    Private IntegralNew As Double
    Private IntegralOld As Double
    Private DerivateNew As Double
    Private ControllerOut As Double

    Sub ResetTable()
        ReDim Table(1, -1)
        Enabled = False
    End Sub

    Sub AddTableLine(ByVal RawValue As Single, ByVal Temperature As Single)
        Nvalues = Table.GetLength(1)
        ReDim Preserve Table(1, Nvalues)
        Table(0, Nvalues) = RawValue
        Table(1, Nvalues) = Temperature
        Nvalues += 1
        If Nvalues >= 2 Then Enabled = True
    End Sub

    Function InterpolateRawValue(ByVal x As Double) As Double
        Dim i As Int32 = 1
        While i < Nvalues - 1 AndAlso Table(0, i) < x
            i = i + 1
        End While
        InterpolateRawValue = Table(1, i - 1) + _
                              ((x - Table(0, i - 1)) / _
                              (Table(0, i) - Table(0, i - 1))) * _
                              (Table(1, i) - Table(1, i - 1))
    End Function

    Function GetText() As String
        If Not Enabled Then Return " - - - "
        Return Measured.ToString("0.0", Globalization.CultureInfo.InvariantCulture).PadLeft(5)
    End Function

    Sub SetDesiredTempToDefault()
        IntegralOld = 0
        DesiredTemp = DefaultTemp
        Slots.WriteSlot(SlotHeater, 0)
    End Sub

    Sub ControlTemperature()
        ' -------------------------------------------------------------
        If Not Enabled Then Return
        If Table.Length < 1 Then Return
        ' ------------------------------------------------------------- Error New
        Measured = Slots.ReadSlot_NoNan(SlotSensor)
        Measured = InterpolateRawValue(Measured)
        ErrorNew = DesiredTemp - Measured
        ' ------------------------------------------------------------- Integral term
        IntegralNew = IntegralOld + _
                      Igain * ErrorNew * LoopTime
        ' ------------------------------------------------------------- Derivative term
        DerivateNew = (ErrorNew - ComputeDerivativeTerm()) / LoopTime
        ' ------------------------------------------------------------- Controller out
        ControllerOut = Pgain * ErrorNew + _
                        Dgain * ComputeDerivativeTerm() + _
                        IntegralNew
        ' ------------------------------------------------------------- Limit output values and
        ' ------------------------------------------------------------- Integral wind up protection
        If ControllerOut < 0 Then
            ControllerOut = 0
            IntegralNew = IntegralOld
        End If
        If ControllerOut > 1000 Then
            ControllerOut = 1000
            IntegralNew = IntegralOld
        End If
        ' ------------------------------------------------------------- Write to slot
        Slots.WriteSlot(SlotHeater, ControllerOut)
        ' ------------------------------------------------------------- Old values
        IntegralOld = IntegralNew
        ErrorOld = ErrorNew
        Fifo1.InsertValue(ErrorNew)
        ' ------------------------------------------------------------- Debug
        'Form1.Text = (Pgain * ErrorNew).ToString("0.0", GCI) + " " + _
        '              IntegralNew.ToString("0.0", GCI) + " " + _
        '              (Dgain * ComputeDerivativeTerm()).ToString("0.0", GCI)
    End Sub

    ' ------------------------------------------------------------------------
    ' Temperature is controlled every 100 mS
    ' this FIFO increases the time to 1000 mS to stabilze the derivative value
    ' ------------------------------------------------------------------------
    Private Fifo1 As FifoDouble = New FifoDouble(10)

    Private Function ComputeDerivativeTerm() As Double
        Return (ErrorNew - Fifo1.OldestValue) / Fifo1.Capacity
    End Function

End Class


Public Module TemperatureControlFunctions
    ' ==================================================================================================
    '   TEMPERATURES
    ' ==================================================================================================
    Friend ExtruderSelected As Int32 = 1

    Friend Sub CNC_SetSelectedExtruderTemperature(ByVal value As String)
        Dim temp As Single = CSng(Val(value))
        If ExtruderSelected = 1 Then TempExtruder1.DesiredTemp = temp
        If ExtruderSelected = 2 Then TempExtruder2.DesiredTemp = temp
    End Sub

    Private TempStopWatch As Stopwatch = New Stopwatch

    Friend Sub CNC_ControlAllTemperatures()
        If Not TemperatureController.Active Then Return
        ' -------------------------------------------------------------------- start the stop watch
        If Not TempStopWatch.IsRunning Then TempStopWatch.Start()
        ' -------------------------------------------------------------------- wait a loop time of 100 mS
        TemperatureController.LoopTime = TempStopWatch.Elapsed.TotalSeconds
        If TemperatureController.LoopTime > 0.1 Then
            TempStopWatch.Reset()
            TempStopWatch.Start()
            ' ------------------------------------------------- temperature controllers
            TempAmbient.ControlTemperature()
            TempChamber.ControlTemperature()
            TempPrintBed.ControlTemperature()
            TempExtruder1.ControlTemperature()
            TempExtruder2.ControlTemperature()
            ' ------------------------------------------------- wait extruder temperature
            If CNC_WaitingTemperature Then
                Dim deltatemp As Double
                If ExtruderSelected = 1 Then deltatemp = TempExtruder1.DesiredTemp - TempExtruder1.Measured
                If ExtruderSelected = 2 Then deltatemp = TempExtruder2.DesiredTemp - TempExtruder2.Measured
                If Math.Abs(deltatemp) < 0.2 Then
                    CNC_WaitingTemperature = False
                End If
            End If
        End If
    End Sub

    Friend Sub AllTemperaturesToDefault()
        TempAmbient.SetDesiredTempToDefault()
        TempChamber.SetDesiredTempToDefault()
        TempPrintBed.SetDesiredTempToDefault()
        TempExtruder1.SetDesiredTempToDefault()
        TempExtruder2.SetDesiredTempToDefault()
    End Sub

End Module